"""Custom exceptions for PyPrefab."""

import typing as t
from gettext import gettext as _

import click
import typer


def style_message(message) -> str:
    """
    Style the message for display.

    Use Click's style function to format the error message because error
    messages emitted from the callback function of an interactive prompt
    use Click's echo function, which does not support rich text.
    https://click.palletsprojects.com/en/stable/api/#click.style
    """
    message = click.style(f"{message}")  # colors would go here
    message = f"❌ {message}"
    return message


class PyprefabBadParameter(typer.BadParameter):
    """Custom exception for bad parameters in PyPrefab CLI."""

    def __init__(self, message: str):
        super().__init__(message)
        self.message = style_message(message)
        self.show_color = True

    def show(self, file: t.Optional[t.IO[t.Any]] = None) -> None:
        """
        Override show() method in Click exceptions.

        This method is currently a no-op, because Click does not
        use the show() method to print exception messages when the
        exception comes from prompted input.

        The related Click code is in termui.prompt(). To override the default
        behavior, you can change the except UsageError block in prompt_func()
        to:

         except UsageError as e:
            if hide_input:
                echo(_("Error: The value you entered was invalid."), err=err)
            else:
               # use the exception's show() method to print the error message
               e.show()
        """
        click.echo(_(self.message), err=True)



"""Command-line interface for the pyprefab package."""

import os
import shutil
import sys
import traceback
from datetime import datetime
from pathlib import Path
from typing import Optional

import structlog
import typer
from jinja2 import Environment, FileSystemLoader, select_autoescape
from rich import print
from rich.console import Console
from rich.panel import Panel
from rich.theme import Theme
from typing_extensions import Annotated

from pyprefab import __version__
from pyprefab.exceptions import PyprefabBadParameter

logger = structlog.get_logger()

cli_theme = Theme(
    {
        "help": "bold cyan",
        "option": "bold yellow",
        "argument": "bold magenta",
    }
)

# Create a console with the custom theme
console = Console(theme=cli_theme)
app = typer.Typer(
    add_completion=False,
    help="Generate python package scaffolding based on pyprefab.",
    rich_markup_mode="rich",
)


def validate_package_name(value: str) -> str:
    """Validate that package name follows Python package naming conventions."""
    if not value.isidentifier():
        if value[0].isdigit():
            msg = "Python package names cannot start with a number"
        else:
            msg = "Python package names must contain letters, numbers, or underscores"
        raise PyprefabBadParameter(msg)
    else:
        return value


def validate_author_desc(value: str) -> str:
    """Validate package author and description."""
    if '"' in value:
        msg = 'Author and description cannot contain double quotes (")'
        raise PyprefabBadParameter(msg)
    else:
        return value


def validate_package_dir(value: Path) -> Path:
    """Validate the target directory of the new package."""
    # use os.path instead of pathlib for the next two checks because Windows
    # Pathlib objects don't have expanduser and resolve methods
    target_dir_str = os.path.expanduser(value)
    target_dir = Path(os.path.normpath(target_dir_str))

    if target_dir.is_file():
        raise PyprefabBadParameter(f"{str(target_dir)} is a file, not a directory")

    # Target directory should be empty (with a few exceptions)
    allow_existing = [".git"]
    exceptions = [allow for allow in allow_existing if (target_dir / allow).is_dir()]
    if target_dir.exists() and sum(1 for item in target_dir.iterdir()) - len(exceptions) > 0:
        raise PyprefabBadParameter(f"{str(target_dir)} is not an empty directory")

    return target_dir


def version_callback(value: bool):
    """Return the package version."""
    if value:
        print(f"pyprefab {__version__}")
        raise typer.Exit()


def render_templates(context: dict, templates_dir: Path, target_dir: Path):
    """Render Jinja templates to target directory."""
    # Process templates
    env = Environment(
        loader=FileSystemLoader(templates_dir),
        trim_blocks=True,
        lstrip_blocks=True,
        keep_trailing_newline=True,
        autoescape=select_autoescape(
            default=True,
        ),
    )
    # For rendering path names
    path_env = Environment(
        trim_blocks=True,
        lstrip_blocks=True,
        keep_trailing_newline=True,
        autoescape=select_autoescape(
            default=True,
        )
    )

    for template_file in templates_dir.rglob("*"):
        if template_file.is_file():
            rel_path = template_file.relative_to(templates_dir)
            if str(rel_path.parents[0]).startswith("docs") and not context.get("docs"):
                continue
            template = env.get_template(str(rel_path))
            output = template.render(**context)

            # Process path parts through Jinja
            path_parts = []
            for part in rel_path.parts:
                # Render each path component through Jinja
                logger.debug("rendering jinja template", template=part)
                rendered_part = path_env.from_string(part).render(**context)
                if rendered_part.endswith(".j2"):
                    rendered_part = rendered_part[:-3]
                path_parts.append(rendered_part)

            # Create destination path preserving structure
            dest_file = target_dir.joinpath(*path_parts)
            dest_file.parent.mkdir(parents=True, exist_ok=True)
            with open(dest_file, "w", encoding="utf-8", newline="\n") as f:
                f.write(output)


@app.command()
def main(
    name: Annotated[
        str,
        typer.Option(
            help="Name of the package",
            prompt=typer.style("Package name 🐍", fg=typer.colors.MAGENTA, bold=True),
            callback=validate_package_name,
            show_default=False,
        ),
    ],
    author: Annotated[
        Optional[str],
        typer.Option(
            help="Package author",
            prompt=typer.style("Package author 👤", fg=typer.colors.MAGENTA, bold=True),
            callback=validate_author_desc,
            show_default=False,
        ),
    ] = "None",
    description: Annotated[
        Optional[str],
        typer.Option(
            help="Package description",
            prompt=typer.style("Package description 📝", fg=typer.colors.MAGENTA, bold=True),
            callback=validate_author_desc,
            show_default=False,
        ),
    ] = "None",
    package_dir: Annotated[
        Path,
        typer.Option(
            "--dir",
            help="Directory that will contain the package",
            prompt=typer.style("Package directory 🎬", fg=typer.colors.MAGENTA, bold=True),
            show_default="current directory",
            callback=validate_package_dir,
        ),
    ] = Path.cwd(),
    docs: Annotated[
        Optional[bool],
        typer.Option(
            "--docs",
            help="Include Sphinx documentation files",
            prompt=typer.style("Include Sphinx docs? 📄", fg=typer.colors.MAGENTA, bold=True),
        ),
    ] = False,
    version: Annotated[
        Optional[bool],
        typer.Option("--version", help="Show the version and exit", callback=version_callback, is_eager=True),
    ] = None,
):
    """
    Pyprefab generates the "get started quickly" scaffolding for a new
    Python package.

    To get started, type "pyprefab" and follow the prompts.

    Full documentation: https://bsweger.github.io/pyprefab/

    """
    templates_dir = Path(__file__).parent / "templates"
    target_dir = package_dir

    current_year = datetime.now().year

    logger.debug("creating package", package=name, directory=str(target_dir))

    try:
        # Create package directory
        target_dir.mkdir(parents=True, exist_ok=True)
        # Template context
        context = {
            "author": author,
            "current_year": current_year,
            "description": description,
            "docs": docs,
            "package_name": name,
        }

        # Write Jinja templates to package directory
        render_templates(context, templates_dir, target_dir)
        panel_msg = (
            f"✨ Created new package [bold green]{name}[/] in {target_dir}\n"
            f"Author: [blue]{author}[/]\n"
            f"Description: {description}"
        )
        if docs:
            panel_msg += f"\nDocumentation: {target_dir}/docs"
        print(
            Panel.fit(
                panel_msg,
                title="Package Created Successfully",
                border_style="green",
            )
        )

    except Exception as e:
        err_console = Console(stderr=True)
        err_console.print(
            Panel.fit(
                f"⛔️ Error creating package: {str(e)}",
                title="pyprefab error",
                border_style="red",
            )
        )
        typer.secho(f"Error creating package: {str(e)}", fg=typer.colors.RED)
        if target_dir.exists():
            shutil.rmtree(target_dir)
        logger.debug(
            "package creation failed",
            package=name,
            directory=str(target_dir),
            error=str(e),
            traceback=traceback.format_exc(),
        )
        raise typer.Exit(1)

    logger.debug("package created", package=name, directory=str(target_dir))


if __name__ == "__main__":
    sys.exit(app())  # pragma: no cover



"""Test functionality of a Python package created by pyprefab."""

import ast
import subprocess


def test_build(cli_output):
    """Files created should build a valid python package."""
    package_path, cli_result = cli_output

    # pyprefab output should be a valid Python package
    result = subprocess.run(
        ["python", "-m", "build", "--sdist", "--wheel"], capture_output=True, cwd=package_path, text=True
    )
    assert result.returncode == 0
    assert "deprecationwarning" not in result.stdout.lower()


def test_lint(cli_output):
    """Files created by CLI should lint without errors."""
    package_path, cli_result = cli_output

    result = subprocess.run(["ruff", "check", package_path], capture_output=True, text=True)
    assert result.returncode == 0


def test_format(cli_output):
    """Files created by CLI should meet formatting standards."""
    package_path, cli_result = cli_output
    result = subprocess.run(["ruff", "format", "--check", package_path], capture_output=True, text=True)
    assert result.returncode == 0


def test_logging(cli_output):
    """Generated Python package should have functional logging."""
    package_path, cli_result = cli_output

    module_path = package_path / "src"
    with open(module_path / "log_test.py", "w", encoding="utf-8") as f:
        f.write("from transporter_logs import app\n")
        f.write('app.logger.info("log test")\n')

    result = subprocess.run(["python", "log_test.py"], capture_output=True, cwd=module_path, text=True)
    logs = ast.literal_eval(result.stdout)
    assert logs.get("event").lower() == "log test"
    assert logs.get("level").lower() == "info"



import logging
import os
import tomllib  # type: ignore
from pathlib import Path
from typing import Any


class PyprefabConfig:
    """
    Manages pyprefab package configuration.

    This class loads package configuration from a TOML file and also applies
    any environment variable overrides.
    """

    def __init__(self, env_prefix: str = "PYPREFAB"):
        self.env_prefix = env_prefix.upper()
        self._config: dict[str, Any] = {}

        self._load_config()
        self.validate_config()

    def _load_config(self) -> None:
        """
        Load config with precedence order:
        1. TOML file (lowest)
        2. Environment variables (highest)
        """
        self._load_config_toml()
        self._apply_env_overrides()

    def _load_config_toml(self) -> None:
        """Load pyprefab .toml config."""
        config_path = Path(__file__).parent / "config.toml"

        if config_path.exists():
            with open(config_path, "rb") as f:
                self._config = tomllib.load(f)
        else:
            self._config = {}

    def _apply_env_overrides(self) -> None:
        """
        Apply environment variable overrides.

        Environment variable format: {PREFIX}_{SECTION}_{KEY}
        Example: PYPREFAB_LOGGING_LEVEL overrides config['logging']['level']
        """
        prefix = f"{self.env_prefix}_"
        env_overrides = {key[len(prefix) :]: value for key, value in os.environ.items() if key.startswith(prefix)}

        if not env_overrides:
            return

        for env_key, env_value in env_overrides.items():
            config_path = [part.lower() for part in env_key.split("_")]

            # set config value
            self._set_nested_config_value(config_path, env_value)

    def _set_nested_config_value(self, config_path: list[str], env_value: str) -> None:
        """
        Update self._config with a nested value from a config path.

        Args:
            config_path: List of keys representing the path in the config
            env_value: String value to set at the deepest level

        Example:
            _set_nested_config_value(['logging', 'level'], 'DEBUG')
            updates self._config['logging']['level'] = 'DEBUG'
        """
        if not config_path:
            return

        current = self._config
        for key in config_path[:-1]:
            if key not in current:
                current[key] = {}
            elif not isinstance(current[key], dict):
                current[key] = {}
            current = current[key]

        current[config_path[-1]] = env_value

    def get_package_setting(self, key: str, default=None):
        """Get pyprefab package settings."""
        keys = key.split(".")
        value = self._config
        for k in keys:
            if isinstance(value, dict) and k in value:
                value = value[k]
            else:
                return default
        return value

    def validate_config(self):
        """Validate the loaded configuration."""
        if not isinstance(self._config, dict):
            raise ValueError("Configuration must be a dictionary.")

        log_level = self.get_package_setting("logging.level")

        # check for valid logging level
        if log_level is not None and not hasattr(logging, log_level):
            raise ValueError(f"Invalid logging level: {log_level}")

        return True



"""pyprefab initialization."""

import sys
from importlib.metadata import PackageNotFoundError, version

from pyprefab.config import PyprefabConfig
from pyprefab.logger import configure_logging

# get application config
try:
    CONFIG = PyprefabConfig()
except ValueError as e:
    sys.exit(f"Invalid pyprefab configuration: {e}")

# configure structlog
configure_logging(CONFIG)

try:
    __version__ = version("pyprefab")
except PackageNotFoundError:  # pragma: no cover
    # package is not installed
    pass



"""Test the pyprefab cli."""

import os
from pathlib import Path
from unittest.mock import patch

import pytest
from click import unstyle
from typer.testing import CliRunner

import pyprefab
from pyprefab.cli import (
    app,  # type: ignore
    validate_author_desc,
    validate_package_name,
)
from pyprefab.exceptions import PyprefabBadParameter


@pytest.mark.parametrize(
    "package_name, expected_result",
    [
        ("valid_package_name", "valid_package_name"),
        ("valid_package_name123", "valid_package_name123"),
        ("AnnoyingButValid", "AnnoyingButValid"),
    ],
)
def test_valid_package_name(package_name, expected_result):
    """Test package name validation."""
    assert validate_package_name(package_name) == expected_result


@pytest.mark.parametrize(
    "package_name",
    [
        ("invalid-package-name"),
        ("invalid-package-name!"),
        ("1invalid_package_name"),
    ],
)
def test_invalid_package_name(package_name):
    """Test package name validation."""
    with pytest.raises(PyprefabBadParameter):
        validate_package_name(package_name)


@pytest.mark.parametrize(
    "author_name",
    [
        ('"double quote" not allowed"'),
        ('"'),
    ],
)
def test_invalid_author(author_name):
    """Test author name and package description name validation."""
    with pytest.raises(PyprefabBadParameter):
        validate_author_desc(author_name)


@pytest.mark.parametrize(
    "cli_inputs, expected_dirs",
    [
        (["--name", "pytest_package", "--author", "Py Test"], [".github", "src", "test"]),
        (["--name", "pytest_package", "--author", "Py Test", "--docs"], [".github", "docs", "src", "test"]),
    ],
)
def test_pyprefab_cli(tmp_path, cli_inputs, expected_dirs):
    runner = CliRunner()
    package_dir = tmp_path / "test_cli"

    result = runner.invoke(
        app,
        cli_inputs + ["--dir", package_dir],
        input="''\nn\n",
    )
    assert result.exit_code == 0
    assert package_dir.exists()

    # package directory populated with template output contains expected folders
    dir_count = 0
    dir_names = []
    for child in package_dir.iterdir():
        if child.is_dir():
            dir_names.append(child.name)
            dir_count += 1
    assert dir_count == len(expected_dirs)
    assert set(dir_names) == set(expected_dirs)


def test_version_param():
    """Version param runs eagerly and exits."""
    runner = CliRunner()
    result = runner.invoke(
        app,
        ["--name", "bad-package-name!", "--version"],
    )
    assert result.exit_code == 0
    assert unstyle(result.stdout).strip() == f"pyprefab {pyprefab.__version__}"


def test_app_invalid_package_name(tmp_path):
    """Package name must be a valid Python identifier."""
    runner = CliRunner()
    result = runner.invoke(
        app,
        ["--name", "pytest-package", "--author", "Py Test", "--dir", tmp_path],
        input="This is a test package\n",
    )
    assert result.exit_code != 0


def test_app_package_dir_callback(monkeypatch, tmp_path):
    """App should fail if supplied package_dir is reserved."""
    # override the current home directory (for testing ~, etc.)
    envs = {"HOME": str(tmp_path), "USERPROFILE": str(tmp_path)}
    monkeypatch.setattr(os, "environ", envs)

    package_dir = "~/test_package_dir"
    runner = CliRunner()
    # boolean prompts require explicit 'n' or 'y'
    result = runner.invoke(
        app, ["--name", "pytest_package", "--author", "Py Test", "--dir", package_dir], input="''\nn\n"
    )
    assert result.exit_code == 0


def test_app_package_dir_file(monkeypatch, tmp_path):
    """App should fail if supplied package_dir is a file."""
    # override the current home directory (for testing ~, etc.)
    envs = {"HOME": str(tmp_path), "USERPROFILE": str(tmp_path)}
    monkeypatch.setattr(os, "environ", envs)

    # set up test by creating an existing file
    package_dir = "~/../test_package_dir//textfile.txt"
    normalized_dir = Path(package_dir).expanduser().resolve()
    normalized_dir.parent.mkdir(parents=True, exist_ok=True)
    normalized_dir.touch()

    runner = CliRunner()
    result = runner.invoke(
        app,
        ["--name", "pytest_package", "--author", "Py Test", "--dir", package_dir],
        input="This is a test package\n",
    )
    assert result.exit_code != 0


def test_error_cleanup(tmp_path):
    """Error when creating a package should trigger cleanup."""
    package_dir = tmp_path / "test_error"
    with patch("pyprefab.cli.render_templates", side_effect=Exception("Test exception")):
        runner = CliRunner()
        result = runner.invoke(
            app,
            ["--name", "pytest_package", "--description", "", "--dir", package_dir],
            input="Py Test\n",
        )
    assert result.exit_code != 0
    # pyprefab should remove package directory if an error occurs
    assert package_dir.exists() is False


def test_error_existing_data(tmp_path):
    """If there are existing files in package directory, pyprefab should fail."""

    package_dir = tmp_path / "test_existing_data"
    package_dir.mkdir()
    (package_dir / "existing_file.txt").touch()

    runner = CliRunner()
    result = runner.invoke(
        app,
        ["--name", "pytest_package", "--author", "Py Test", "--description", "new package", "--dir", package_dir],
        input="y\n",
    )
    assert result.exit_code != 0


def test_existing_data_exception(tmp_path):
    """If existing files in package directory are on the exception list, pyprefab should create the package."""

    package_dir = tmp_path / "test_existing_data"
    (package_dir / ".git").mkdir(parents=True, exist_ok=True)

    runner = CliRunner()
    result = runner.invoke(
        app,
        ["--name", "pytest_package", "--author", "Py Test", "--description", "new package", "--dir", package_dir],
        input="y\n",
    )
    assert result.exit_code == 0


def test_existing_data_exception_and_no_exception(tmp_path):
    """If there's a mix of allowed and not allowed existing files in the package directory, pyprefab should fail."""

    package_dir = tmp_path / "test_existing_data"
    (package_dir / ".git").mkdir(parents=True, exist_ok=True)
    (package_dir / "logs").mkdir(parents=True, exist_ok=True)

    runner = CliRunner()
    result = runner.invoke(
        app,
        ["--name", "pytest_package", "--author", "Py Test", "--description", "new package", "--dir", package_dir],
        input="y\n",
    )
    assert result.exit_code != 0


@pytest.mark.parametrize(
    "cli_inputs, expected_author",
    [
        (["--name", "pytest_package", "--author", "Authorin'' with many 'names 🧐", "--description", "testy"], "Authorin'' with many 'names 🧐"),
        (["--name", "pytest_package", "--author", "Author, Name", "--description", "testy"], "Author, Name"),
        (["--name", "pytest_package", "--author", "'", "--description", "testy"], "'"),
        (["--name", "pytest_package", "--author", "''", "--description", "testy"], "''"),
    ],
)
def test_pyprefab_author_names(tmp_path, cli_inputs, expected_author):
    """Author name should not be auto-escaped in jinja templates."""
    package_name = "pytest_package"
    runner = CliRunner()
    package_path = tmp_path / package_name
    result = runner.invoke(
        app,
        cli_inputs + ["--dir", package_path],
        input="y\n",
    )
    assert result.exit_code == 0
    with open(package_path / "pyproject.toml", "r", encoding="utf-8") as f:
        file = f.read()
        assert f'authors = [{{name = "{expected_author}"}}]' in file


@pytest.mark.parametrize(
    "cli_inputs, expected_description",
    [
        (["--name", "pytest_package", "--author", "Worf", "--description", "app for trackin' things"], "app for trackin' things"),
        (["--name", "pytest_package", "--author", "Worf", "--description", "''''"], "''''"),
        (["--name", "pytest_package", "--author", "Worf", "--description", "look ma, usin' an em dash: ⎯"], "look ma, usin' an em dash: ⎯"),
    ],
)

def test_pyprefab_description(tmp_path, cli_inputs, expected_description):
    """Author name should not be auto-escaped in jinja templates."""
    package_name = "pytest_package"
    runner = CliRunner()
    package_path = tmp_path / package_name
    result = runner.invoke(
        app,
        cli_inputs + ["--dir", package_path],
        input="y\n",
    )
    assert result.exit_code == 0
    with open(package_path / "pyproject.toml", "r", encoding="utf-8") as f:
        file = f.read()
        assert f'description = "{expected_description}"' in file



"""Snapshot tests for pyprefab templates."""

import tomllib  # type: ignore

import pytest


def test_pyproject_docs(cli_output, snapshot):
    """pyproject.toml contents are correct for project with docs."""
    package_path, cli_result = cli_output
    with open(package_path / "pyproject.toml", "rb") as f:
        pyproject = tomllib.load(f)
    assert pyproject.get("dependency-groups", {}).get("docs")
    assert pyproject == snapshot


@pytest.mark.parametrize(
    "meta_file",
    [
        "CHANGELOG.md",
        "CONTRIBUTING.md",
        "LICENSE",
        "README.md",
    ],
)
def test_meta_files(cli_output, snapshot, meta_file):
    """Location and contents of CHANGELOG.md are correct."""
    package_path, cli_result = cli_output
    with open(package_path / meta_file, "r", encoding="utf-8") as f:
        file_contents = f.read()
    assert file_contents == snapshot


@pytest.mark.parametrize(
    "docs_file",
    [
        "conf.py",
        "index.rst",
        "CHANGELOG.md",
        "CONTRIBUTING.md",
        "readme.md",
        "usage.md",
    ],
)
def test_docs_dir(cli_output, snapshot, docs_file):
    """Documentation contents are correct."""
    package_path, cli_result = cli_output
    with open(package_path / "docs" / "source" / docs_file, "r", encoding="utf-8") as f:
        file = f.read()
    assert file == snapshot


def test_pyproject_no_docs(cli_output_no_docs, snapshot):
    """pyproject.toml contents are correct for project without docs."""
    package_path, cli_result = cli_output_no_docs
    with open(package_path / "pyproject.toml", "rb") as f:
        pyproject = tomllib.load(f)
    assert pyproject.get("dependency-groups", {}).get("docs") is None
    assert pyproject == snapshot


@pytest.mark.parametrize(
    "src_file",
    [
        "__init__.py",
        "app.py",
        "logger.py",
    ],
)
def test_src_dir(cli_output, snapshot, src_file):
    """Template files in src/ rendered correctly."""
    package_path, cli_result = cli_output
    with open(package_path / "src" / "transporter_logs" / src_file, "r", encoding="utf-8") as f:
        file = f.read()
    assert file == snapshot


@pytest.mark.parametrize(
    "gh_workflow_file",
    [
        "ci.yaml",
        "publish-pypi.yaml",
    ],
)
def test_gh_workflows(cli_output, snapshot, gh_workflow_file):
    """Github workflow templates rendered correctly."""
    package_path, cli_result = cli_output
    with open(package_path / ".github" / "workflows" / gh_workflow_file, "r", encoding="utf-8") as f:
        file = f.read()
    assert file == snapshot



"""Test pyprefab app configuration."""

import pytest

from pyprefab.config import PyprefabConfig


@pytest.fixture
def setenv(monkeypatch):
    """Fixture to set environment variables."""
    monkeypatch.setenv("PYPREFAB_TESTY_LEVELS_GREETING", "HOWDY")
    monkeypatch.setenv("PYPREFAB_LOGGING_LEVEL", "DEBUG")
    yield


def test_config_prefix():
    config = PyprefabConfig()
    assert isinstance(config._config, dict)
    assert config.env_prefix == "PYPREFAB"


def test_config_custom_prefix():
    config = PyprefabConfig(env_prefix="HOWDY")
    assert config.env_prefix == "HOWDY"


def test_config_env_override(setenv):
    config = PyprefabConfig()
    assert config.get_package_setting("testy.levels.greeting") == "HOWDY"
    assert config.get_package_setting("testy.levels") == {"greeting": "HOWDY"}
    assert config.get_package_setting("testy") == {"levels": {"greeting": "HOWDY"}}
    assert config.get_package_setting("logging.level") == "DEBUG"
    assert config.get_package_setting("fakekey.mwahaha") is None


def test_validate_config(monkeypatch):
    """Invalid logging level should raise a config error."""
    monkeypatch.setenv("PYPREFAB_LOGGING_LEVEL", "INVALID_LEVEL")
    with pytest.raises(ValueError, match="Invalid logging level"):
        PyprefabConfig()



import logging
import sys

import structlog


def configure_logging(config) -> None:
    # Set logging level from config if available
    log_level = config.get_package_setting("logging.level")
    if log_level is not None:
        logging.root.setLevel(log_level)

    shared_processors = [
        structlog.processors.TimeStamper(fmt="iso", utc=True),
        structlog.processors.add_log_level,
        structlog.processors.CallsiteParameterAdder([structlog.processors.CallsiteParameter.FILENAME]),
    ]

    if sys.stderr.isatty():
        # If we're in a terminal, pretty print the logs.
        processors = shared_processors + [
            structlog.dev.ConsoleRenderer(),
        ]  # pragma: no cover
    else:
        # Otherwise, output logs in JSON format
        processors = shared_processors + [
            structlog.processors.dict_tracebacks,
            structlog.processors.JSONRenderer(),
        ]

    structlog.configure(
        processors=processors,
        wrapper_class=structlog.make_filtering_bound_logger(logging.root.level),
        cache_logger_on_first_use=True,
    )



# Configuration file for the Sphinx documentation builder.
#
# For the full list of built-in configuration values, see the documentation:
# https://www.sphinx-doc.org/en/master/usage/configuration.html

# -- Project information -----------------------------------------------------
# https://www.sphinx-doc.org/en/master/usage/configuration.html#project-information

import os
import sys
from datetime import datetime
from importlib import metadata

# If extensions (or modules to document with autodoc) are in another directory,
# add these directories to sys.path here. If the directory is relative to the
# documentation root, use os.path.abspath to make it absolute, like shown here.
sys.path.insert(0, os.path.abspath("..."))
sys.path.insert(0, os.path.abspath(".."))
sys.path.insert(0, os.path.abspath("."))

# -- General configuration ---------------------------------------------------
# https://www.sphinx-doc.org/en/master/usage/configuration.html#general-configuration

extensions = [
    "sphinx.ext.autodoc",
    "sphinx.ext.githubpages",
    "sphinx.ext.intersphinx",
    "sphinx.ext.napoleon",
    "sphinx.ext.viewcode",
    "sphinx_copybutton",
    "sphinxcontrib.typer",
    "sphinxext.opengraph",
    "myst_parser",
]

myst_enable_extensions = [
    "colon_fence",
    "fieldlist",
    "replacements",
    "tasklist",
]

# Settings from the copybutton's config
# https://github.com/executablebooks/sphinx-copybutton/blob/master/docs/conf.py
# There are useful for ensuring that the copy button only copies code
# (as prefixed by ">>>" for example) and not the prompt or the output.
copybutton_prompt_text = r">>> |\.\.\. |\$ |In \[\d*\]: | {2,5}\.\.\.: | {5,8}: "
copybutton_prompt_is_regexp = True
copybutton_line_continuation_character = "\\"
copybutton_here_doc_delimiter = "EOT"
copybutton_selector = "div:not(.no-copybutton) > div.highlight > pre"

# Open graph settings
ogp_site_url = "https://bsweger.github.io/pyprefab/"
ogp_image = "https://bsweger.github.io/pyprefab/_static/pyprefab.png"

templates_path = ["_templates"]
exclude_patterns = []

# The suffix(es) of source filenames.
# You can specify multiple suffix as a list of string:
source_suffix = [".rst", ".md"]

# The encoding of source files.
# source_encoding = 'utf-8-sig'

# The master toctree document.
master_doc = "index"

# project information
now = datetime.now()
project = "pyprefab"
copyright = f"{now.year}, Becky Sweger | Last update {now.strftime('%B %d, %Y')}"
author = "Becky Sweger"

version = metadata.version("pyprefab")
release = version

# The language for content autogenerated by Sphinx. Refer to documentation
# for a list of supported languages.
language = "en"

# List of patterns, relative to source directory, that match files and
# directories to ignore when looking for source files.
exclude_patterns = ["_build"]

# The name of the Pygments (syntax highlighting) style to use.
pygments_style = "monokai"

# If true, `todo` and `todoList` produce output, else they produce nothing.
todo_include_todos = False


# -- Options for HTML output -------------------------------------------------
# https://www.sphinx-doc.org/en/master/usage/configuration.html#options-for-html-output

html_theme = "alabaster"
html_static_path = ["_static"]

# Theme options are theme-specific and customize the look and feel of a theme
# further. For a list of options available for each theme, see the
# documentation.
html_theme_options = {
    "description": f"Small CLI for Python project scaffolding\n{release}",
    "github_user": "bsweger",
    "github_repo": "pyprefab",
    "github_banner": True,
    "github_button": False,
    "font_family": "'Roboto', Georgia, sans",
    "head_font_family": "'Roboto', Georgia, serif",
    "code_font_family": "'Roboto Mono', 'Consolas', monospace",
    "pre_bg": "#433e56",
}

# Add any paths that contain custom themes here, relative to this directory.
# html_theme_path = []

# The name for this set of Sphinx documents.  If None, it defaults to
# "<project> v<release> documentation".
# html_title = None

# A shorter title for the navigation bar.  Default is the same as html_title.
# html_short_title = None

# The name of an image file (relative to this directory) to place at the top
# of the sidebar.
# html_logo = None

# The name of an image file (within the static path) to use as favicon of the
# docs.  This file should be a Windows icon file (.ico) being 16x16 or 32x32
# pixels large.
# html_favicon = None

# Add any paths that contain custom static files (such as style sheets) here,
# relative to this directory. They are copied after the builtin static files,
# so a file named "default.css" will overwrite the builtin "default.css".
html_static_path = ["_static"]

# These paths are either relative to html_static_path
# or fully qualified paths (eg. https://...)
html_css_files = [
    "custom.css",
]

# Add any extra paths that contain custom files (such as robots.txt or
# .htaccess) here, relative to this directory. These files are copied
# directly to the root of the documentation.
# html_extra_path = []

# If not '', a 'Last updated on:' timestamp is inserted at every page bottom,
# using the given strftime format.
html_last_updated_fmt = "%b %d, %Y"

# If true, SmartyPants will be used to convert quotes and dashes to
# typographically correct entities.
# html_use_smartypants = True

# Custom sidebar templates, maps document names to template names.
# html_sidebars = {
#     '**': [
#         'about.html',
#         'navigation.html',
#         'relations.html',
#         'searchbox.html',
#     ]
# }

# Additional templates that should be rendered to pages, maps page names to
# template names.
# html_additional_pages = {}

# If false, no module index is generated.
# html_domain_indices = True

# If false, no index is generated.
# html_use_index = True

# If true, the index is split into individual pages for each letter.
# html_split_index = False

# If true, links to the reST sources are added to the pages.
# html_show_sourcelink = True

# If true, "Created using Sphinx" is shown in the HTML footer. Default is True.
html_show_sphinx = False

# If true, "(C) Copyright ..." is shown in the HTML footer. Default is True.
# html_show_copyright = True

# If true, an OpenSearch description file will be output, and all pages will
# contain a <link> tag referring to it.  The value of this option must be the
# base URL from which the finished HTML is served.
# html_use_opensearch = ''

# This is the file name suffix for HTML files (e.g. ".xhtml").
# html_file_suffix = None

# Language to be used for generating the HTML full-text search index.
# Sphinx supports the following languages:
#   'da', 'de', 'en', 'es', 'fi', 'fr', 'hu', 'it', 'ja'
#   'nl', 'no', 'pt', 'ro', 'ru', 'sv', 'tr'
# html_search_language = 'en'

# A dictionary with options for the search language support, empty by default.
# Now only 'ja' uses this config value
# html_search_options = {'type': 'default'}

# Output file base name for HTML help builder.
# htmlhelp_basename = ""



"""Test pyprefab logger configuration."""

import logging

import pytest
import structlog

from pyprefab.config import PyprefabConfig
from pyprefab.logger import configure_logging


@pytest.fixture
def reset_logging(monkeypatch):
    """Reset logging configuration before and after each test."""
    # clear existing logging level env var and reset to default of WARNING
    monkeypatch.delenv("PYPREFAB_LOGGING_LEVEL", raising=False)
    logging.root.setLevel(logging.WARNING)

    yield

    # clean up after test
    logging.root.setLevel(logging.WARNING)
    structlog.reset_defaults()


def test_logging_level_from_config(reset_logging, monkeypatch):
    """Test that logging level is set from config when specified."""
    monkeypatch.setenv("PYPREFAB_LOGGING_LEVEL", "DEBUG")
    config = PyprefabConfig()
    configure_logging(config)

    assert logging.root.level == logging.DEBUG


def test_logging_level_default_when_not_specified(reset_logging, monkeypatch):
    """Test that logging level uses Python's default WARNING when not specified."""
    monkeypatch.delenv("PYPREFAB_LOGGING_LEVEL", raising=False)
    config = PyprefabConfig()
    config._config = {}
    configure_logging(config)
    assert logging.root.level == logging.WARNING



import pytest
from typer.testing import CliRunner

from pyprefab.cli import app  # type: ignore


@pytest.fixture
def cli_output(tmp_path):
    runner = CliRunner()
    result = runner.invoke(
        app,
        [
            "--name",
            "transporter_logs",
            "--author",
            "Miles O'Brien",
            "--description",
            "An app for parsin' transporter logs",
            "--dir",
            tmp_path,
            "--docs",
        ],
    )
    return tmp_path, result


@pytest.fixture
def cli_output_no_docs(tmp_path):
    runner = CliRunner()
    result = runner.invoke(
        app,
        [
            "--name",
            "transporter_logs",
            "--author",
            "Miles O'Brien",
            "--description",
            "An app for parsin' transporter logs",
            "--dir",
            tmp_path,
        ],
        input="n\n",
    )
    return tmp_path, result



import nox

nox.options.default_venv_backend = "uv"

PYPROJECT = nox.project.load_toml("pyproject.toml")
PYTHON_VERSIONS = nox.project.python_versions(PYPROJECT)
# assume that the programming version classifies in pyproject.toml are in ascending order
PYTHON_LATEST = PYTHON_VERSIONS[-1]
COVERAGE_THRESHOLD = 90


@nox.session(tags=["checks", "ci"])
def lint(session: nox.Session) -> None:
    """Run the linter."""
    session.install("ruff")
    session.run("ruff", "check", ".")


@nox.session(tags=["checks", "ci"])
def typecheck(session: nox.Session) -> None:
    """Run static type checking."""
    session.install("mypy")
    session.run("mypy", "src", "test")


@nox.session(python=PYTHON_VERSIONS, tags=["checks", "ci"])
def tests(session: nox.Session) -> None:
    """Install dependencies from lockfile and run tests."""
    session.run_install(
        "uv",
        "sync",
        "--active",
        "--group=dev",
        "--frozen",
        "--quiet",
        f"--python={session.virtualenv.location}",
    )
    session.run("pytest")


@nox.session(tags=["checks", "ci"], python=PYTHON_LATEST, requires=[f"tests-{PYTHON_LATEST}"])
def coverage(session) -> None:
    """Check test coverage."""
    session.run_install(
        "uv",
        "sync",
        "--active",
        "--group=dev",
        "--frozen",
        "--quiet",
        f"--python={session.virtualenv.location}",
    )
    session.run("coverage", "erase", silent=True)
    session.run("coverage", "run", "-m", "pytest", "--quiet", "--no-header", "--no-summary", silent=True)
    session.run("coverage", "html", "--skip-covered", "--skip-empty", silent=True)
    session.run(
        "coverage",
        "report",
        "--format",
        "markdown",
        f"--fail-under={COVERAGE_THRESHOLD}",
    )


@nox.session(python=PYTHON_VERSIONS)
def test_dev_install(session: nox.Session) -> None:
    """Install as a package and run tests."""
    session.run_install(
        "uv",
        "sync",
        "--active",
        "--group=dev",
    )
    session.run("pytest")


@nox.session(python=PYTHON_VERSIONS)
def test_pypi_install(session: nox.Session) -> None:
    """Install as a package and run tests."""
    session.run_install("uv", "pip", "install", "pyprefab")


@nox.session(python=PYTHON_LATEST, tags=["checks", "ci"])
def docs(session: nox.Session) -> None:
    """Build the documentation."""
    session.run_install(
        "uv",
        "sync",
        "--active",
        "--group=docs",
        "--frozen",
        "--quiet",
        f"--python={session.virtualenv.location}",
    )
    session.run("sphinx-build", "-W", "-b", "html", "docs/source", "docs/_build/html")


@nox.session(default=False)
def docs_serve(session: nox.Session) -> None:
    """Serve the documentation locally."""
    session.run_install(
        "uv",
        "sync",
        "--active",
        "--group=docs",
        "--frozen",
        "--quiet",
        f"--python={session.virtualenv.location}",
    )
    session.run("sphinx-autobuild", "docs/source", "docs/_build/html", external=True)



"""Test pyprefab exceptions."""

from click import unstyle

from pyprefab.exceptions import PyprefabBadParameter


def test_pyprefab_bad_parameter_show(capsys):
    """PyprefabBadParameter show method works as expected."""
    pyprefab_exception = PyprefabBadParameter("oops!")
    assert pyprefab_exception.show_color is True
    assert unstyle(pyprefab_exception.message) == "❌ oops!"
    pyprefab_exception.show()
    captured = capsys.readouterr()
    assert captured.err.strip() == "❌ oops!"
